package fi.jyu.mit.ohj2;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;
import java.util.TreeMap;


/**
 * Luokka ini-tiedoston ksittelyyn.  Tiedoston muoto:
 * <pre>
 * [FormGUI]
 * leveys=200
 * korkeus=100
 * cbEtsi=1
 * [Positions]
 * FormGUI=0,0
 * </pre>
 * 
 * Usea ohjelman osa voi "luoda" tiedoston uudestaan, koska jo avattut
 * tiedostot puskuroidaan, eik niit avata en toista kertaa.
 * Tm ei kuitenkaan ole viel thread-safe.
 * 
 * @author vesal
 * @version 14.6.2013
 * 
 * @example
 * <pre name="test">
 * #import java.io.File;
 *   File fini = new File("testini.ini");
 *   File fini2 = new File("testini2.ini");
 *   fini.delete();  fini2.delete();
 *   
 *   IniFile ini = IniFile.create("testini.ini");
 *   ini.write("FormGUI","leveys",200);
 *   ini.write("FormGUI","korkeus",100);
 *   ini.write("FormGUI","cbEtsi",1);
 *   ini.write("Positions","FormGUI","20,30");
 *   
 *   fini.renameTo(fini2) === true; // koska saman luominen ei loisi mitn uutta
 *   
 *   ini = IniFile.create("testini2.ini");
 *   ini.read("FormGUI","leveys",0) === 200;
 *   ini.read("FormGUI","korkeus",0) === 100;
 *   ini.read("FormGUI","cbEtsi",0) === 1;
 *   ini.read("FormGUI","cbEiOo",2) === 2;
 *   ini.read("eioo","cbEiOo",2) === 2;
 *   ini.read("Positions","FormGUI","1,1") === "20,30";
 *   ini.write("FormGUI","cbEtsi",0);
 *   ini.read("FormGUI","cbEtsi",9) === 0;
 *   fini2.delete() === true;
 * </pre>
 * 
 */
public class IniFile  {

    private static TreeMap<String,IniFile> iniFiles = new TreeMap<>();
    
    /**
     * Luodaan uusi ini-file tai annetaan entinen jos on jo kerran luettu
     * @param fileName mink nimist tiedostoa ksitelln
     * @return luota tai puskurista otettu tiedosto
     */
    public static IniFile create(String fileName) {
        File file = new File(fileName);
        String fullname = file.getAbsolutePath();
        IniFile result = iniFiles.get(fullname);
        if ( result != null ) return result;
        result = new IniFile(fullname);
        iniFiles.put(fullname, result);
        return result;
    }
    
    private String fileName;
    private TreeMap<String,TreeMap<String,String>> sections = new TreeMap<>(); 
    
    /**
     * @param fileName luettavan tiedoston nimi
     */
    private IniFile(String fileName) {
        this.fileName = fileName;
        readFile();
    }

    
    private String lastSectionName = null;
    private TreeMap<String,String> lastSection = null; 
    
    
    private TreeMap<String, String> getSectionValue(String section) {
        if ( section.equals(lastSectionName) ) return lastSection;
        lastSectionName = section;
        lastSection = sections.get(section);
        if ( lastSection != null ) return lastSection;
        lastSection = new TreeMap<>();
        sections.put(section, lastSection);
        return lastSection;
    }


    private TreeMap<String, String> getSectionValueNoCreate(String section) {
        if ( section.equals(lastSectionName) ) return lastSection;
        TreeMap<String,String> last = sections.get(section);
        if ( last == null ) return null;
        lastSection = last;
        lastSectionName = section;
        return lastSection;
    }


    private void writeNoSave(String section, String item, String value) {
        TreeMap<String,String> sectionValue = getSectionValue(section);
        sectionValue.put(item, value.replaceAll("\n", "\\\\n"));
    }
    
    
    private void readFile() {
        String sectionName = "";
        try ( BufferedReader fi = new BufferedReader(new FileReader(fileName)) ) {
            String rivi;
            while ((rivi = fi.readLine()) != null) { 
                rivi = rivi.trim();
                if ( rivi.startsWith("[")) {
                    sectionName = rivi.replace('[', ' ').replace(']', ' ').trim();
                    continue;
                }
                int p = rivi.indexOf('=');
                if ( p < 0 ) continue;
                String itemName = rivi.substring(0,p);
                String value = rivi.substring(p+1);
                writeNoSave(sectionName, itemName, value);
            }

        } catch ( FileNotFoundException e ) {
            // Mits tehtis
        } catch (IOException e) {
            // Mits tehtis
        }
    }
    
    
    /**
     * Tallenttaa koko rakenteen
     */
    public void saveFile()  {
        File ftied = new File(fileName);

        try ( PrintWriter fo = new PrintWriter(new FileWriter(ftied.getCanonicalPath())) ) {
            for (Map.Entry<String,TreeMap<String,String>> entry : sections.entrySet()) {
                fo.println("[" + entry.getKey() + "]");
                for (Map.Entry<String,String> item : entry.getValue().entrySet()) {
                    fo.println(item.getKey() + "=" + item.getValue());
                }
            }
        } catch ( FileNotFoundException ex ) {
            // Mits tehtis
        } catch ( IOException ex ) {
            // Mits tehtis
        }

    }


    /**
     * Kirjoitetaan uusi arvo 
     * @param section mihin osioon kirjoitetaan
     * @param item mik on nimi
     * @param value mik on uusi arvo
     */
    public void write(String section, String item, String value) {
        writeNoSave(section, item, value);
        saveFile();
    }
    
    
    /**
     * Kirjoitetaan uusi arvo 
     * @param section mihin osioon kirjoitetaan
     * @param item mik on nimi
     * @param value mik on uusi arvo
     */
    public void write(String section, String item, int value) {
        write(section,item,""+value);
    }
    
    
    /**
     * Luetaan arvo
     * @param section mist osioista luetaan
     * @param item mist kohdasta luetaan
     * @param defValue mik arvo palautetaan jos itemia ei lydy
     * @return lydetty arvo tai oletus
     */
    public String read(String section, String item, String defValue) {
        TreeMap<String,String> sectionValue = getSectionValueNoCreate(section);
        if ( sectionValue == null ) return defValue;
        String result = sectionValue.get(item);
        if ( result == null ) return defValue;
        result = result.replaceAll("\\\\n", "\n");
        return result;
    }
    
    
    /**
     * Luetaan arvo
     * @param section mist osioista luetaan
     * @param item mist kohdasta luetaan
     * @return lydetty arvo tai null jos ei lydy
     */
    public String read(String section, String item) {
        return read(section,item,(String)null);
    }
    
    
    /**
     * Luetaan arvo
     * @param section mist osioista luetaan
     * @param item mist kohdasta luetaan
     * @param defValue mik arvo palautetaan jos itemia ei lydy
     * @return lydetty arvo tai oletus
     */
    public int read(String section, String item, int defValue) {
        String result = read(section,item);
        if ( result == null ) return defValue;
        return Mjonot.erotaInt(result,defValue);
    }


}
