package fi.jyu.mit.ohj2;
import java.io.*;
import java.util.*;

/**
 * Luokka avustusten tulostamiseksi.  Avustustiedoston muoto:
 * <pre>
 * [SISÄLLYS] - aina etsitään sisällystä tällä aihe- (topic) nimellä.
 * Eka aihe - kerrotaan ekasta aiheesta
 * Toka aihe - kerrotaan tokasta aiheesta
 *
 * [Eka aihe]
 * Ekassa aiheessa voidaan kuvata mitä vaan ekaan aiheeseen liittyvää
 * ja miksei muutakin.
 *
 * [Toka aihe] - aiheen otsikkorivillä saa olla kommentti
 * Tokassa aiheessa sitten tokan aiheen sisällöstä.
 * Tietysti aiheita voidaan kirjoittaa niin monta kuin halutaan
 * eikä niiden kaikkien tarvitse olla sisällysluettelossa.
 * Jos on kamalan pitkä aihe, jonka tulostus halutaan keskeyttää,
 * voidaan avustustiedoston rivi aloittaa
 * #
 * risuaitamerkillä, joka pysäyttää tulostuksen.
 * muutenkin tulostetaan vain korkeintaan 23 riviä kerrallaan.
 * ; puolipisteellä alkava rivi on kommenttia ja sitä ei tulosteta
 * </pre>
 * Käyttöesimerkkejä:<br>
 * Selaillaan koko avustusta.  Tulostetaan ensin sisällysluettelo
 * <pre>
 *   try {
 *     Help h = new Help("kerho.hlp");
 *     h.browse();
 *   }
 *   catch (IOException ioe) {
 *     System.err.println(ioe);
 *   }
 * </pre>
 * Tulostetaan valitun aiheen kohdalta:
 * <pre>
 *   h.printMatchingTopics("Li*");  // tulostaa kaikki Li-alkavat aiheet
 *   h.printTopic("Lisäys");        // tulostaa aiheen Lisäys
 *   h.helpTopic("Li*");            // tulostaa kaikki Li-alkavat aiheet
 * </pre>
 * @author Vesa Lappalainen, Markku Vire
 * @version 1.0, 24.03.2003
 */
public class Help {
  private static final char COMMENT = ';';
  private static final char TOPIC_START = '[';
  private static final char TOPIC_END = ']';
  private static final char HELP_PAUSE = '#';
  private static final char QUIT_CHAR = 'q';
  private static final int  DEFAULT_LINES = 23;
  private static final String INDEX = "SIS*LLYS";

  private final Map<String,Collection<String>> topics = new HashMap<String,Collection<String>>();
  private int lines = DEFAULT_LINES;
  private int printedLineCounter = 0;
  private final BufferedReader consoleIn = new BufferedReader(new InputStreamReader(System.in));
  
  private PrintStream out = System.out;
  private boolean stdout = true;

  /**
   * @return kuinka monta riviä tulostetaan korkeintaan pysähtymättä
   */
  public int getLines() { return lines; }

  /**
   * @param lines pysähtymättä tulostettavien rivien lukumäärä
   */
  public void setLines(int lines) {
    if ( lines > 0 )
      this.lines = lines;
  }

  /**
   * @param input tutkittava rivi
   * @return alkaako rivi lopetusmerkillä
   */
  private static boolean quit(String input) {
    return firstIs(input, QUIT_CHAR);
  }

  /**
   * @param s tutkittava rivi
   * @param c merkki jota etsitään
   * @return onko c rivin s ensimmäinen merkki (true) vai ei (false)
   */
  private static boolean firstIs(String s, char c) {
    if ( s == null || s.length() == 0 )
      return false;

    return (Character.toUpperCase(s.charAt(0)) == Character.toUpperCase(c));
  }

  /**
   * Pysähtyy odottamaan returnin panamista jos on tulostettuja rivejä
   * edellisen pysähtymisen jälkeen.
   * @return onko painettu postumista (true) vai ei (false)
   */
  private boolean helpStop() {
    if ( !stdout ) return false;  
    String input = null;

    if (printedLineCounter != 0) {
      printedLineCounter = 0;
      System.out.print("Jatka ENTER > ");
      try {
        input = consoleIn.readLine();
      } catch (IOException ioe) {
        System.out.println(ioe);
      }
    }

    return quit(input);
  }

  /**
   * Lisätään uusi aihe-otsikko avustukseen.
   * Tämän jälkeen rivejä voi lisätä otsikon alle
   * <pre>
   * Collection topic = h.addTopic("Uusi aihe");
   * topic.add("Eka rivi");
   * topic.add("Toka rivi");
   * </pre>
   * Mikäli aihe on jo olemassa, palautetaan viite vanhaan aiheeseen.
   * @param topic lisättävän aiheen otsikko
   * @return tietorakenne, johon voi lisätä rivejä aiheen alle
   */
  public final Collection<String> addTopic(String topic) {
    String uptopic = topic.toUpperCase(); 
    Collection<String> newtopic = topics.get(uptopic);
    if ( newtopic != null ) return newtopic;

    newtopic = new ArrayList<String>();
    topics.put(uptopic,newtopic);
    return newtopic;
  }

  /**
   * Lisätään yksi rivi avustukseen aiheen topic alle.  Mikäli aihetta
   * ei vielä ole, se luodaan.
   * @param topic lisättävän aiheen otsikko
   * @param line lisättävä rivi
   */
  public void addLine(String topic, String line) {
    Collection<String> topicLines = addTopic(topic);
    topicLines.add(line);
  }

  /**
   * Alustetaan tyhjä avustus, johon voi lisätä aiheita
   * metodeilla: addTopic, addLine, readFile.
   */
  public Help()  { 
      // ei tarvii tehdä mitään toistaiseksi
  }
  
  
  /**
   * Asetetaan tulostusvirta toiseen paikkaan
   * @param newout mihin tulostus tehdään?
   */
  public void setOut(PrintStream newout) {
      out = newout;
      stdout = false;
  }
  
  

  /**
   * Lukee avustuksen tiedostosta.  Voidaan kutsua useita kertoja
   * jolloin voidaan yhdistää monia avustustiedostoja.
   *
   * Tiedot kerätään map-tauluun, jossa aiheiden mukaiset merkkijonot
   * avaimina (isoksi muutettuna).  Aiheen alaiset rivit on
   * sitten vektorina "avaimen oliona".
   * @param fileName tiedosto,josta avustukset luetaan
   * @throws IOException jos jokin menee pieleen tiedoston lukemisessa
   */
  public final void readFile(String fileName) throws IOException {
    @SuppressWarnings("resource")
    BufferedReader in = new BufferedReader(new FileReader(fileName));
    Collection<String> topic = null;
    String line;
    try {

      while ( (line = in.readLine()) != null ) {
        int comment = line.indexOf(COMMENT);

        if ( comment >= 0 ) line = line.substring(0, comment);

        if ( firstIs(line, TOPIC_START) ) {
          int topicEnd = line.indexOf(TOPIC_END);

          if (topicEnd >= 0) {
            String topicName = line.substring(1, topicEnd);

            topic = addTopic(topicName);
          }
        } else {
          if (topic != null)
            topic.add(line);
        }
      }
    } finally {
      in.close();
    }
  }

  /**
   * Alustaa avustuksen lukemalla avustukset tiedostosta.
   * Tiedot kerätään map-tauluun, jossa aiheiden mukaiset merkkijonot
   * avaimina (isoksi muutettuna).  Aiheen alaiset rivit on
   * sitten vektorina "avaimen oliona".
   * @param fileName tiedosto,josta avustukset luetaan
   * @throws IOException jos jokin menee pieleen tiedoston lukemisessa
   */
  public Help(String fileName) throws IOException {
    readFile(fileName);
  }

  /**
   * Tulostaa seuraavan rivin Help-tekstiä.  Jos rivimäärä ylittyy,
   * niin pysähtyy.
   * @param text tulostettava rivi
   * @return painettiinko poistumista (true) vai ei (false) tulostuksen aikana
   */
  private boolean helpPrint(String text) {
    if ( firstIs(text, HELP_PAUSE) ) return helpStop();

    out.println(text);
    printedLineCounter++;

    if ( printedLineCounter >= lines ) return helpStop();

    return false;
  }

  /**
   * Tulostaa valitun lohkon avustuksesta.
   * @param topic tulostettava lohko.  Ei saa sisältää jokereita
   * @return painettiinko poistumista (true) vai ei (false) tulostuksen aikana
   */
  public boolean printTopic(String topic) {
    Collection<String> topicText = topics.get(topic.toUpperCase()); 

    if ( topicText == null )
      return helpPrint("Aihetta " + topic + " ei löydy");

    for (Iterator<String> i = topicText.iterator(); i.hasNext(); )
      if ( helpPrint(i.next()) )
        return true;

    return false;
  }

  /**
   * Tulostaa ne avustuksen lohkot jotka täsmäävät hakuehtoon.
   * @param topic mahdolisesti jokereita * ja ? sisältävä ehto
   * @return painettiinko poistumista (true) vai ei (false) tulostuksen aikana
   */
  public boolean printMatchingTopics(String topic) {
    printedLineCounter = 0;

    if ( WildChars.containsWildChars(topic) ) {
      for ( Iterator<String> i = topics.keySet().iterator(); i.hasNext() ; ) {
        String s = i.next();
        if ( WildChars.onkoSamat(s, topic) && printTopic(s) )
          return true;
      }
      return false;
    }
    return printTopic(topic);
  }

  /**
   * Selailee avustusta valitun lohkon kohdalta.  Jos lohko == null
   * niin näytetään kohta [SISÄLLYS]
   * @param topic tulostettava avustuksen kohta
   */
  public void browse(String topic)  {
    String newtopic = topic;  
    while (true) {
      try {
        if (newtopic == null) newtopic = INDEX;

        if ( printMatchingTopics(newtopic) || !stdout ) return;

        System.out.print("Valitse aihe (voi olla myös *) > ");
        newtopic = consoleIn.readLine();
        if ( newtopic.length() == 0 ) break;
        if ( quit(newtopic) ) break;
      }
      catch (IOException ioe) {
        System.out.println(ioe);
      }
    }
  }

  /**
   * Selailee avustusta aloittaen kohdasta [SISÄLLYS]
   */
  public void browse()  { browse(null); }

  /**
   * Tulostaa topic:in mukaisen lohkon avustuksesta.
   * @param topic tulostetavan lohkon otsikko tai null jolloin selailee avustusta
   */
  public void helpTopic(String topic)  {
    if ( topic == null ) browse();
    else printMatchingTopics(topic);
  }


  /**
   * Testataa Help-luokkaa
   * @param args ei käytössä
   */
  public static void main(String[] args) {
    try {
      Help h = new Help(); //"kerho.txt");
      h.helpPrint("Terve!");
      h.helpTopic("Lisäys");
      h.helpPrint("#");
      h.browse();
      h.addLine("Uusi otsikko","Eka rivi");
      h.addLine("Uusi otsikko","Toka rivi");
      h.printMatchingTopics("Uusi*");
    }
    catch (Exception ioe) {
      System.err.println(ioe);
    }
  }
}
